/**
   CertRequest.java

   Copyright (C) 1999, Claymore Systems, Inc.
   All Rights Reserved.

   ekr@rtfm.com  Sat Jul 24 19:06:19 1999
*/

package COM.claymoresystems.cert;

import COM.claymoresystems.util.Util;
import COM.claymoresystems.ptls.LoadProviders;
import COM.claymoresystems.crypto.EAYEncryptedPrivateKey;
import COM.claymoresystems.cert.EAYDSAPrivateKey;
import COM.claymoresystems.cert.DERUtils;
import COM.claymoresystems.cert.X509DSAPublicKey;
import COM.claymoresystems.cert.X509RSAPublicKey;
import java.io.*;
import java.security.KeyPairGenerator;
import java.security.interfaces.DSAKeyPairGenerator;
import java.security.SecureRandom;
import java.security.KeyPair;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.interfaces.DSAPrivateKey;
import java.security.interfaces.DSAPublicKey;
import xjava.security.interfaces.CryptixRSAPublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import cryptix.util.mime.Base64OutputStream;

public class CertRequest {
     static {
       LoadProviders.init();
     }

     /** Generate a key pair

         @param type DSA or RSA
         @param size the length
         @param password the password to use to encrypted the key
         @param keyfile the keyfile to store the key in
         @param newParams generate new parameters if using DSA--by default Sun uses fixed precomputed params
         @return the keypair

         @exception NoSuchAlgorithmException if you choose a key we don't know about
         @exception NoSuchProviderException internal errors
         @exception IOException encoding errors
     */
     static KeyPair generateKey(String type,int size,
       String password,BufferedWriter keyfile,boolean newParams) 
       throws NoSuchAlgorithmException, NoSuchProviderException, IOException {
       KeyPairGenerator kg=KeyPairGenerator.getInstance(type);
       PrivateKey priv;
       
       kg.initialize(size);

       if(type.equals("DSA")){
         DSAKeyPairGenerator dsaGen=(DSAKeyPairGenerator)kg;
         dsaGen.initialize(size,newParams,new SecureRandom());
       }
       else{
         kg.initialize(size);
       }
       KeyPair pair=kg.generateKeyPair();

       // We need to map the DSA keys given us by Sun into
       // our own DSA keys because Sun gives us a bogus
       // DER encoding
       if(type.equals("DSA")){
	 priv=new EAYDSAPrivateKey((DSAPrivateKey)pair.getPrivate());
       }
       else {
	 priv=pair.getPrivate();
       }
       
       EAYEncryptedPrivateKey.writePrivateKey(priv,password.getBytes(),keyfile);
       return pair;
     }

     /** Make a Netscape Signed Public Key and Cert request

      @param p the keypair to make it with
      @return the SPKAC as a bytestring
      @exception IOException for errors */
     public static byte[] makeSPKACRequest(KeyPair p)
       throws IOException {
       PrivateKey priv=p.getPrivate();
       PublicKey pub=p.getPublic();
       String alg=priv.getAlgorithm();
       byte[] DSA_algid={(byte)0x30,(byte)0x0b,(byte)0x06,(byte)0x07,(byte)0x2a,
			 (byte)0x86,(byte)0x48,(byte)0xce,(byte)0x38,(byte)0x04,
			 (byte)0x03,(byte)0x05,(byte)0x00};		
       byte[] RSA_algid={(byte)0x30,(byte)0x0d,(byte)0x06,(byte)0x09,
                         (byte)0x2a,(byte)0x86,(byte)0x48,(byte)0x86,(byte)0xf7,
                         (byte)0x0d,(byte)0x01,(byte)0x01,(byte)0x05,(byte)0x05,
                         (byte)0x00};
       
       // We need to map the keys given us by Sun or Cryptix into
       // our own DSA keys because they gives us a bogus
       // DER encoding (or none at all)
       byte[] algId;
       String algorithm;
       
       if(alg.equals("DSA")){
         algId=DSA_algid;
         algorithm="DSA";
	 pub=new X509DSAPublicKey((DSAPublicKey)pub);
       }
       else if(alg.equals("RSA")){
         algId=RSA_algid;
         algorithm="SHA-1/RSA/PKCS#1";
         pub=new X509RSAPublicKey((CryptixRSAPublicKey)pub);
       }
       else throw new InternalError("Unknown algorithm "+alg);

       byte[] key_enc=pub.getEncoded(); // SPKI

       ByteArrayOutputStream os=new ByteArrayOutputStream();
       os.write(key_enc);
       DERUtils.encodeIA5String("Challenge",os);
       byte[] tmp=os.toByteArray();
       os.reset();
       DERUtils.encodeSequence(tmp,os);

       // At this point we have the public key and challenge sequence
       byte[] toBeSigned=os.toByteArray();
       byte[] signature;
       
       try {
	 Signature sig=Signature.getInstance(algorithm);
	 sig.initSign(priv);

	 sig.update(toBeSigned);
	 signature=sig.sign();
       } catch (Exception e){
	 throw new InternalError(e.toString());
       }

       os.reset();
       os.write(toBeSigned);
       // This is a hack. We assume that we're using DSA
       os.write(algId);
       DERUtils.encodeBitString(signature,os);

       byte[] spkac_c=os.toByteArray();
       // Now we've got the SPKAC as a byte array, but we need to wrap it in a sequence
       os.reset();
       DERUtils.encodeSequence(spkac_c,os);

       byte[] spkac=os.toByteArray();
       
       return spkac;
     }

     public static void main(String[] args)
       throws IOException, Exception {
       String keyFileName=args[0];
       String keyType;
       if(args.length==2){
         keyType=args[1];
       }
       else
         keyType="DSA";
       int size=1024;

       InputStreamReader ir=new InputStreamReader(System.in);
       BufferedReader br=new BufferedReader(ir);

       String password=br.readLine();
       
       FileWriter fw=new FileWriter(keyFileName);
       BufferedWriter bw=new BufferedWriter(fw);

       KeyPair kp=generateKey(keyType,size,password,bw,true);
       byte[] spkac=makeSPKACRequest(kp);

       ByteArrayOutputStream bos=new ByteArrayOutputStream();
       Base64OutputStream b64os=new Base64OutputStream(bos);
       b64os.write(spkac);
       b64os.flush();
       b64os.close();
       
       ByteArrayInputStream bis=new ByteArrayInputStream(
	    bos.toByteArray());
       InputStreamReader irr=new InputStreamReader(bis);
       BufferedReader r=new BufferedReader(irr);

       String line;

       System.out.println(password);
       while((line=r.readLine())!=null){
	 System.out.println(line);
       }
     }
}
     

